from collections import Counter
from haystack import Document
from haystack.components.readers import ExtractiveReader
from typing import List, Optional
import os
from .pdf_docx_reader import PDF_DOCX_Reader
import logging

#smaller model: deepset/roberta-base-squad2
class DocumentFilter:
    
    __slots__ = ("_reader", "threshold", "max_per_doc", "top_k", "_documents", "_initialized", "logger")
    
    _initialized: bool
    _reader: ExtractiveReader
    threshold: float
    max_per_doc: int
    top_k: int
    _documents: List[Document]
    
    def __init__(self, model_name="deepset/tinyroberta-squad2", path : Optional[str] = None, threshold : float = 0.4, max_per_doc : int = 2, top_answer : int = 4) -> None:
        #check type first
        self.__enforce_type(threshold, float, "threshold")
        self.__enforce_type(max_per_doc, int, "max_per_doc")
        self.__enforce_type(top_answer, int, "top_answer")
        
        # init not lock
        super().__setattr__("_initialized", False)
        # set core reader
        super().__setattr__("_reader", ExtractiveReader(model=model_name))
        self._reader.warm_up()
        # lock down, reader can't be modified
        super().__setattr__("_initialized", True)
    
        self.threshold = threshold
        self.max_per_doc = max_per_doc
        self.top_k = top_answer
        self._documents: List[Document] = [] #init var to hold document
        if path is None:
            path = os.getcwd()
        if not os.path.exists(path):
            raise FileNotFoundError(f"The path '{path}' does not exist.")
        
        self.logger = logging.getLogger(self.__class__.__name__)
        self.logger.propagate = False  # Prevent propagation to the root logger
        handler = logging.StreamHandler()
        handler.setFormatter(logging.Formatter(
            '[%(asctime)s] - [%(name)s] - [%(levelname)s] - %(message)s',
            datefmt='%Y-%m-%d %H:%M:%S'
        ))
        self.logger.addHandler(handler)
        self.logger.setLevel(logging.INFO)
        
        self.__init_documents(path) #init documents from the path
        self.logger.info(f"Initialize successfully at path {path}")
    
    @property
    def reader(self):
        return self._reader
    
    @property
    def documents(self):
        return self._documents
       
    def __setattr__(self, name, value):
        # once initialized, prevent changing core internals
        if getattr(self, "_initialized", False) and name in ("_reader"):
            raise AttributeError(f"Cannot reassign '{name}' after initialization")
        super().__setattr__(name, value)
    
    def search_document(self, prompt : str = None) -> List:
        self.__enforce_type(prompt, str, "prompt") #check type before start
        
        result = self._reader.run(query=prompt, documents=self._documents, top_k=self.top_k)

        seen_texts = set()
        counts = Counter()
        filtered_answers = []

        for ans in result["answers"]:
            if ans.score < self.threshold:
                continue
            if ans.document is None:
                continue
            text = ans.document.content
            doc_id = ans.document.id

            if counts[doc_id] >= self.max_per_doc:
                continue
            if text in seen_texts:
                continue

            filtered_answers.append((text, ans.score, doc_id))
            seen_texts.add(text)
            counts[doc_id] += 1

        #contain text, score, and doc.id
        return filtered_answers #return list of document with ranking score that > threshold

    def __init_documents(self, directory : str = "") -> None:
        #get from here
        self._documents.clear()
        pdf_urls, docx_urls =self.__collect_file_paths(directory)
        
        reader = PDF_DOCX_Reader() #init reader
        
        #extract from docx
        for doc in docx_urls:
            text = reader.extract_ordered_text(doc)
            self._documents.append(Document(content=text))
            
        #extract from pdf
        for pdf in pdf_urls:
            text = reader.extract_ordered_text(pdf)
            self._documents.append(Document(content=text))
    
    def __collect_file_paths(self, directory):
        pdf_urls = []
        docx_urls = []

        # Walk through all directories and files
        for root, _, files in os.walk(directory):
            for file in files:
                file_path = os.path.join(root, file).replace('\\', '/')  # normalize path
                if file.lower().endswith('.pdf'):
                    pdf_urls.append(file_path)
                elif file.lower().endswith('.docx'):
                    docx_urls.append(file_path)

        return pdf_urls, docx_urls
    
    def __enforce_type(self, value, expected_type, arg_name):
        if not isinstance(value, expected_type):
            raise TypeError(f"Argument '{arg_name}' must be of type {expected_type.__name__}, but received {type(value).__name__}")


from gtts import gTTS, lang # need pip install gTTS
import time
import os
import playsound # need pip install playsound==1.2.2 (the later version wouldn't work)
from multiprocessing.managers import Namespace

def gtts_speak(text: str = None, lang: str = 'vi', shared_name: Namespace = None) -> str:
    if text is None:
        raise ValueError("text could not be None, did you forget what to speak?")
    elif not isinstance(text, str):
        raise TypeError("text must be str")
    temp_filename = str(round(time.time() * 10)) + ".mp3"
    script_dir = os.path.dirname(os.path.abspath(__file__))
    temp_ID_path = os.path.join(script_dir, temp_filename)

    tts = gTTS(text=text, lang=lang)
    tts.save(temp_ID_path)

    if shared_name is not None:
        shared_name.path = temp_ID_path #shared_name.path will store the file path, enabling cleanup if the multiprocessing process is terminated prematurely.

    try:
        playsound.playsound(temp_ID_path)
    finally:
        if os.path.exists(temp_ID_path):
            os.remove(temp_ID_path)

    return temp_ID_path

#for those who need what language it supports and the code lang for it
def gtts_print_supported_languages() -> None:
    supported_languages = lang.tts_langs()  # Get the dictionary of supported languages
    print("Supported languages by gTTS:")
    for code, language in supported_languages.items():
        print(f"{code}: {language}")

#how to use
if __name__=="__main__":
    gtts_print_supported_languages()
    _temp_ID = gtts_speak("Xin chào người đẹp, em tên là gì thế", "vi")
    print("You can get file name as output: "+ _temp_ID)
    
    from multiprocessing import Process, Manager
    _manager = Manager()
    _namespace = _manager.Namespace()
    _namespace.path = ""

    _p = Process(target=gtts_speak, args=("Xin chào bạn, mình là An, hiện là học sinh lớp 12 trường trung học phổ thông chuyên", "vi", _namespace))
    _p.start()
    time.sleep(3)
    _p.terminate()
    _p.join()

    print("Path from shared namespace:", _namespace.path)
    if os.path.exists(_namespace.path):
        os.remove(_namespace.path)
        print("Cleaned up leftover file.")
    else:
        print("No file to clean up.")


#this is just about langdetect library, since they already simple enough, there is no need to make function 

from langdetect import detect, detect_langs #need pip install langdetect

if __name__ == "__main__":
    text1 = "This is a sample text in English. Xin chào bạn đẹp trai "
    detected_language1 = detect(text1)
    print(f"The language of '{text1}' is: {detected_language1}")

    text2 = "Ceci est un exemple de texte en français."
    detected_language2 = detect_langs(text2) #it return a list
    print(f"The language of '{text2}' is: {detected_language2}")

    text3 = "Đây là một đoạn văn bản mẫu bằng tiếng Việt."
    detected_language3 = detect_langs(text3)

    if detected_language3:
        most_likely_language = detected_language3[0]
        language_code = most_likely_language.lang
        probability = most_likely_language.prob

        print(f"The detected language of '{text3}' is: {language_code}")
        print(f"The confidence score is: {probability:.4f}")

        extracted_values = [language_code, probability]
        print(f"Extracted values as a list: {extracted_values}")
    else:
        print(f"Could not detect the language of '{text3}'.")


# freeai_utils/__init__.py
import importlib
from typing import TYPE_CHECKING

# all class/functions available
_module_to_names = {
    'audio_record': ['WavRecorder','MP3Recorder','check_wav_length_and_size','check_mp3_length_and_size'],
    'audio_to_text_whisper': ['OpenAIWhisper'],
    'audio_to_text_vn': ['VN_Whisper'],
    'clean_text_for_tts': ['clean_ai_text_for_tts'],
    'cleaner': ['__Cleaner'],
    'document_filter': ['DocumentFilter'],
    'geminiAPI': ['GeminiClient'],
    'google_search': ['GoogleSearcher'],
    'image_to_text': ['ImageCaptioner'],
    'text_from_image_easyocr': ['Text_Extractor_EasyOCR'],
    'text_to_speech_gtts': ['gtts_print_supported_languages', 'gtts_speak'],
    'text_to_speech_pyttsx3': ['Text_To_Speech_Pyttsx3'],
    'pdf_docx_reader': ['PDF_DOCX_Reader'],
    'wrapper': ['time_it'],
}

# make a map from the list: a lookup to import specific files for needed tools, avoiding full library load.
_lazy_mapping = {}
for module_name, symbols in _module_to_names.items():
    for symbol in symbols:
        _lazy_mapping[symbol] = module_name

# all class/functions available when do import freeai_utils
__all__ = sorted(_lazy_mapping.keys())


def __getattr__(name: str):
    if name in _lazy_mapping:
        # Import only the module that contains the requested symbol
        module = importlib.import_module(f"{__name__}.{_lazy_mapping[name]}")
        value = getattr(module, name)
        globals()[name] = value  # Cache in namespace for future calls
        return value
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


def __dir__():
    return __all__


#for IDEs suggestions
if TYPE_CHECKING:
    from .audio_record             import WavRecorder, MP3Recorder, check_wav_length_and_size, check_mp3_length_and_size
    from .audio_to_text_whisper    import OpenAIWhisper
    from .audio_to_text_vn         import VN_Whisper
    from .clean_text_for_tts       import clean_ai_text_for_tts
    from .cleaner                  import __Cleaner
    from .document_filter          import DocumentFilter
    from .geminiAPI                import GeminiClient
    from .google_search            import GoogleSearcher
    from .image_to_text            import ImageCaptioner
    from .text_from_image_easyocr  import Text_Extractor_EasyOCR
    from .text_to_speech_gtts      import gtts_print_supported_languages, gtts_speak
    from .text_to_speech_pyttsx3   import Text_To_Speech_Pyttsx3
    from .pdf_docx_reader          import PDF_DOCX_Reader
    from .wrapper                  import time_it


import os
from typing import List, Tuple

class __Cleaner: #this class first created is due to the gttS doesnt delete the .mp3 100% each time when interrupt in a multiprocessing 
    def __init__(self, directory : str | None = None):
        if directory is None:
            self._dir = os.path.dirname(os.path.abspath(__file__))
        else: self._dir = directory
        
        if type(self._dir) is not str:
            raise TypeError(f"the directory must be str type in order to proceed")
        
        if not os.path.exists(self._dir):
            raise FileNotFoundError(f"Your path doesn't exist")
        
    
    def remove_all_files_end_with(self, ends_with : str = None) -> Tuple[List[str], int]: #return the amount of files just got removed
        counter : int = 0
        paths : List[str] = []
        
        if type(ends_with) is not str or (isinstance(ends_with, str) and not ends_with.startswith('.')) or len(ends_with) < 2:
            raise TypeError(f"The 'ends_with' argument must be a string representing a file extension (e.g., '.txt'). You provided: {ends_with!r}")
        
        for root, _, files in os.walk(self._dir):
            for file in files:
                file_path = os.path.join(root, file).replace('\\', '/')  # normalize path
                if file.lower().endswith(ends_with):
                    counter += 1
                    paths.append(file_path)
                    os.remove(file_path)
        
        return paths, counter            
        
if __name__ == "__main__":
    cleaner = __Cleaner()
    print(cleaner.remove_all_files_end_with('.png'))


import torch
import librosa
from transformers import AutoProcessor, AutoModelForSpeechSeq2Seq
from transformers import logging as transformers_logging
import logging

#other model_id: "namphungdn134/whisper-base-vi"
class VN_Whisper:
    def __init__(self, model_id="namphungdn134/whisper-small-vi", device : str | None = None) -> None:
        self._model = None 
        self._device = None 
        self.logger = logging.getLogger(self.__class__.__name__)
        
        #init the var to hold device available
        preferred_devices = []
        
        #try input first
        if device is not None:
            preferred_devices.append(device)
        
        # try cuda second 
        if torch.cuda.is_available() and "cuda" not in preferred_devices:
            preferred_devices.append("cuda")

        # fall back to CPU if not already there
        if "cpu" not in preferred_devices:
            preferred_devices.append("cpu")
            
        for dev in preferred_devices:
            try:
                self.logger.info(f"Loading '{model_id}' model on {dev}...")
                # Load processor and model
                self.processor = AutoProcessor.from_pretrained(model_id)
                self._model = AutoModelForSpeechSeq2Seq.from_pretrained(model_id).to(dev)
                self.logger.info(f"Model successfully loaded on {dev}.")
                self._device = dev
                break
            except RuntimeError as e:
                self.logger.error(f"Failed to load on {dev}: {e}")
            except Exception as e: 
                self.logger.error(f"An unexpected error occurred while loading model on {dev}: {e}")
        
        if self._model is None: #check if nothing works, then raise error
            raise RuntimeError(f"Could not load model on any device")
        # Suppress transformer-related warnings
        transformers_logging.set_verbosity_error()
    
    @property
    def model(self):
        return self._model  
    
    @property
    def device(self):
        return self._device
              
    def transcribe_audio(self, audio_path) -> str:
        # Load and preprocess audio
        self.logger.info(f"Loading audio from: {audio_path}")
        audio, sr = librosa.load(audio_path, sr=16000)
        input_features = self.processor(audio, sampling_rate=16000, return_tensors="pt").input_features.to(self._device)

        # Generate transcription
        self.logger.info("Generating transcription...")
        with torch.no_grad():
            predicted_ids = self._model.generate(input_features, num_beams=1, use_cache=True)

        # Decode transcription
        transcription = self.processor.batch_decode(predicted_ids, skip_special_tokens=True)[0]
        transcription = self.__verify(transcription)
        return transcription

    def __verify(self, prompt : str) -> str:
        # 1.this to check if nothing is heard
        # 2. that phrase looks like the default model use it as temp phrase when nothing is heard, probably to ensure the audio when heard nothing would return a str
        # 3. this check to make sure it doesnt just get some 1 word speech and translate to that
        if prompt.strip() == "" or prompt.strip() == "thôi để con đo huyết áp trước cái" or len(prompt) < 10: 
            return ""
        else: return prompt



import re 
#made by AI, need to check again later

#Strip out markdown/HTML artifacts so that text-to-speech reads clean prose.
def clean_ai_text_for_tts(text: str) -> str:
    if not isinstance(text, str):
            raise TypeError("Text must be str type.")
    
    # 1. Remove fenced code blocks (```…```)
    text = re.sub(r'```[\s\S]*?```', '', text)
    text = re.sub(r'```[\s\S]*?', '', text)
    
    # 2. Remove inline code (`…`)
    text = re.sub(r'`([^`\n]+)`', r'\1', text)
    
    # 3. Remove Markdown images (![alt](url)) entirely
    text = re.sub(r'!\[.*?\]\(.*?\)', '', text)
    
    # 4. Convert Markdown links [text](url) → text
    text = re.sub(r'\[([^\]]+)\]\([^)]+\)', r'\1', text)
    
    # 5. Remove raw URLs
    text = re.sub(r'https?://\S+', '', text)
    
    # 6. Remove HTML tags
    text = re.sub(r'<[^>]+>', '', text)
    
    # 7. Strip Markdown bold/italic markers
    text = re.sub(r'(\*\*|__)(.*?)\1', r'\2', text)
    text = re.sub(r'(\*|_)(.*?)\1',   r'\2', text)
    
    # 8. Remove Markdown headings (lines starting with one-or-more #)
    text = re.sub(r'^\s{0,3}#{1,6}\s+', '', text, flags=re.MULTILINE)
    
    # 9. Remove list bullets or numbered lists
    text = re.sub(r'^\s*[-*+]\s+', '', text, flags=re.MULTILINE)
    text = re.sub(r'^\s*\d+\.\s+',  '', text, flags=re.MULTILINE)
    
    # 10. Remove Python comments (lines starting with #)
    text = re.sub(r'^\s*#.*$', '', text, flags=re.MULTILINE)
    
    # 11. Collapse multiple blank lines to a single one
    text = re.sub(r'\n{2,}', '\n\n', text)
    
    # 12. Trim leading/trailing whitespace
    return text.strip()

if __name__=="__main__":
    _raw = """
        # Summary

        Heres what you need to do:

        1. Clone the repo: `git clone https://github.com/example/project.git`
        2. Install deps via `pip install -r requirements.txt`
        3. Run `<script>.py`

        ```python
        # this code block will be removed entirely
        print("hello world")
        """
    print(clean_ai_text_for_tts(_raw))


# Library for people who want to do ocr but don't want to download Tesseract OCR
# Weaker than Tesseract OCR

import pyautogui # need pip install pyautogui (use to screenshot)
import easyocr# need pip install easyocr (use to read text from image)
import time
from typing import Optional

class Text_Extractor_EasyOCR:
    def __init__(self, language : str ='en') -> None:
        #check type of the input
        self.__enforce_type(language, str, "language")
        
        self._reader = easyocr.Reader([language])
        self.screen_width, self.screen_height = pyautogui.size()
        self._capture_region = (0, 0, self.screen_width, self.screen_height) # default, capture fullscreen

    @property
    def capture_region(self) -> None:
        return self._capture_region
    
    @capture_region.setter
    def capture_region(self, value: tuple) -> None: #this to modify the region manually
        if not isinstance(value, tuple) or len(value) != 4:
            raise ValueError("capture_region must be a tuple of size 4.")
        self._capture_region = value
    
    #this function asks for an image_path and perform ocr on it
    def get_text_from_image(self, image_path : str = None) -> str:
        self.__enforce_type(image_path, str, "image_path")
        result = self._reader.readtext(image_path, detail=0)
        return " ".join(result)
        
    #this function will take a screenshot at your current screen and return the text from the screenshot
    def get_text_from_screenshot(self, capture_region: Optional[tuple] = None, image_name : Optional[str] = None) -> str:
        __region = self._capture_region
        # Rationale: This design provides flexibility.  
        # The class's `self._capture_region` acts as a default, while the `capture_region` parameter allows for one-off, 
        # specific capture areas without altering the default setting. 
        # This avoids repetitive specification of the same region when capturing multiple screenshots with the default area.
        # image_name: will be used to name the screenshot when use the function
        
        if capture_region is not None: #if default, take base on config
            if len(capture_region) != 4:
                raise ValueError(f"capture_region must have size == 4")
            
            __region = capture_region
            
        #check type of the input
        self.__enforce_type(__region, tuple, "capture_region")
        if image_name is None:
            _temp_ID = (str)(round(time.time() * 10)) + ".png"   
        else: 
            self.__enforce_type(image_name, str, "image_name")
            _temp_ID = image_name
        screenshot = pyautogui.screenshot(region=__region)  #capture
        screenshot.save(_temp_ID) #careful if there is a file screenshot before run, it would replace that file
        result = self._reader.readtext(_temp_ID, detail=0)
        return " ".join(result)  #extract and return str

    #set capture region base on percentage instead of choosing specific size capture
    def set_capture_region(self, crop_left: float = 0, crop_right: float = 0, crop_up: float = 0, crop_down: float = 0) -> tuple:
        #check type of the input
        self.__enforce_type(crop_left, (int, float), "crop_left")
        self.__enforce_type(crop_right, (int, float), "crop_right")
        self.__enforce_type(crop_up, (int, float), "crop_up")
        self.__enforce_type(crop_down, (int, float), "crop_down")
        
         # Validate crop percentage values
        if not (0 <= crop_left <= 100):
            raise ValueError("crop_left must be between 0 and 100")
        if not (0 <= crop_right <= 100):
            raise ValueError("crop_right must be between 0 and 100")
        if not (0 <= crop_up <= 100):
            raise ValueError("crop_up must be between 0 and 100")
        if not (0 <= crop_down <= 100):
            raise ValueError("crop_down must be between 0 and 100")
        
        top_percentage_to_ignore = crop_up / 100
        bottom_percentage_to_ignore = crop_down / 100
        left_percentage_to_ignore = crop_left / 100
        right_percentage_to_ignore = crop_right / 100

        # Height of region to capture
        top_y = int(self.screen_height * top_percentage_to_ignore)
        bottom_y = int(self.screen_height * (1 - bottom_percentage_to_ignore))
        capture_height = bottom_y - top_y

        # The width of the region to capture
        left_x = int(self.screen_width * left_percentage_to_ignore)
        right_x = int(self.screen_width * (1 - right_percentage_to_ignore))
        capture_width = right_x - left_x 

        self._capture_region = (left_x, top_y, capture_width, capture_height)
        return (left_x, top_y, capture_width, capture_height)
        
    def __enforce_type(self, value, expected_types, arg_name):
        if not isinstance(value, expected_types):
            expected_names = [t.__name__ for t in expected_types] if isinstance(expected_types, tuple) else [expected_types.__name__]
            expected_str = ", ".join(expected_names)
            raise TypeError(f"Argument '{arg_name}' must be of type {expected_str}, but received {type(value).__name__}")
        
#SUPPORTED LANG easyocr:
# Abaza	abq
# Adyghe	ady
# Afrikaans	af
# Angika	ang
# Arabic	ar
# Assamese	as
# Avar	ava
# Azerbaijani	az
# Belarusian	be
# Bulgarian	bg
# Bihari	bh
# Bhojpuri	bho
# Bengali	bn
# Bosnian	bs
# Simplified Chinese	ch_sim
# Traditional Chinese	ch_tra
# Chechen	che
# Czech	cs
# Welsh	cy
# Danish	da
# Dargwa	dar
# German	de
# English	en
# Spanish	es
# Estonian	et
# Persian (Farsi)	fa
# French	fr
# Irish	ga
# Goan Konkani	gom
# Hindi	hi
# Croatian	hr
# Hungarian	hu
# Indonesian	id
# Ingush	inh
# Icelandic	is
# Italian	it
# Japanese	ja
# Kabardian	kbd
# Kannada	kn
# Korean	ko
# Kurdish	ku
# Latin	la
# Lak	lbe
# Lezghian	lez
# Lithuanian	lt
# Latvian	lv
# Magahi	mah
# Maithili	mai
# Maori	mi
# Mongolian	mn
# Marathi	mr
# Malay	ms
# Maltese	mt
# Nepali	ne
# Newari	new
# Dutch	nl
# Norwegian	no
# Occitan	oc
# Pali	pi
# Polish	pl
# Portuguese	pt
# Romanian	ro
# Russian	ru
# Serbian (cyrillic)	rs_cyrillic
# Serbian (latin)	rs_latin
# Nagpuri	sck
# Slovak	sk
# Slovenian	sl
# Albanian	sq
# Swedish	sv
# Swahili	sw
# Tamil	ta
# Tabassaran	tab
# Telugu	te
# Thai	th
# Tajik	tjk
# Tagalog	tl
# Turkish	tr
# Uyghur	ug
# Ukranian	uk
# Urdu	ur
# Uzbek	uz
# Vietnamese	vi


import google.generativeai as genai #need pip install google-generativeai
import pyperclip #need pip install pyperclip
from dotenv import load_dotenv #need pip install python-dotenv
import os
import logging
from typing import Optional

logging.basicConfig(level=logging.INFO, format='[%(asctime)s] - [%(name)s] - [%(levelname)s] - %(message)s', datefmt='%Y-%m-%d %H:%M:%S')


class GeminiClient:
    def __init__(self, model_name : str ='models/gemini-2.0-flash-lite', api_key : Optional[str] = None, memories_length : int = 4, limit_word_per_respond : int = 150):
        #check data type
        self.__enforce_type(memories_length, int, "memories_length")
        self.__enforce_type(limit_word_per_respond, int, "limit_word_per_respond")
        
        env_path = os.path.join(os.getcwd(), ".env")
        load_dotenv(env_path) # load if there is env file
        self.__api_key = api_key or os.getenv("GEMINI_API_KEY")

        if not self.__api_key:
            raise ValueError("API key was not provided or found in the 'GEMINI_API_KEY' environment variable.")
        
        #check data type before proceed further into connecting
        self.__enforce_type(model_name, str, "model_name")
        self.__enforce_type(self.__api_key, str, "api_key")
        self.logger = logging.getLogger(self.__class__.__name__)
        try:
            genai.configure(api_key=self.__api_key)
            self.logger.info("Gemini API configured successfully.")
        except Exception as e:
            self.logger.critical(f"Error configuring Gemini API: {e}")
            raise

        self._model_name = model_name
        try:
            self._model = genai.GenerativeModel(self._model_name)
            self.logger.info(f"Successfully initialized model: {self._model_name}") 
            self.logger.info(f"Using endpoint: {self._model.model_name}") 
        except Exception as e:
            self.logger.critical(f"ERROR: Error initializing model {self._model_name}: {e}")
            raise
        
        self._history = []
        self._max_length = memories_length
        self._word_prompt = f"Please remember to answer with less than {limit_word_per_respond} words."
    
    @property
    def model_name(self):
        return self._model_name

    @property
    def gemini_model(self):
        return self._model

    @property
    def history(self):
        # Returns the conversation history as a tuple.
        return tuple(self._history)

    @property
    def max_memory_length(self):
        # Returns the maximum length of the conversation history.
        return self._max_length

    @property
    def word_prompt(self):
        # Returns the word limit prompt.
        return self._word_prompt

    @max_memory_length.setter
    def max_memory_length(self, value) -> None:
        # able to change the max_memory_length
        if isinstance(value, int) and value > 0:
            self._max_length = value
        else:
            raise ValueError("Memory length must be a positive integer.")
    
    def ask(self, prompt : str) -> str:
        #not str, return...
        self.__enforce_type(prompt, str, "prompt")
        
        if not prompt:
            self.logger.warning("WARNING: Empty prompt provided, skipping API call.")
            return None
        
        try:
            response = self._model.generate_content(self._word_prompt + prompt)

            if response.parts:
                answer = response.text
                self.logger.info("INFO: Received response successfully.") # Replaced logging.info
                return answer
            else:
                self.logger.info("No response received. This could be due to safety filters or other reasons.") 
                return ""

        except Exception as e:
            print(f"ERROR: An error occurred during the API call: {e}") # Replaced logging.error
            return None   
        
    def list_models(self) -> None:
        #List model available base on your api_key
        print("Retrieve available models...")
        try:
            print("-" * 40)
            for model in genai.list_models():
                # List model with 'generateContent' only
                if 'generateContent' in model.supported_generation_methods:
                    print(f"Model: {model.name}")
                    print(f"  Name: {model.display_name}")
                    print(f"  Description: {model.description}")
                    print("-" * 40)
        except Exception as e:
            self.logger.error(f"Error retrieving: {e}")
   
    def ask_and_copy_to_clipboard(self, prompt : str) -> str:
        answer = self.ask(prompt)
        if answer:
            try:
                pyperclip.copy(answer)
                self.logger.info("Copy to clipboard the answer")
            except Exception as e:
                self.logger.error(f"CANT COPY TO CLIPBOARD, {e}")
        return answer

    def ask_with_memories(self, prompt : str) -> str: # use if you want to store history to ask again (cap with memories_length)
        line = "Here's our previous conversation:\n"
        if self._history:
            for turn in self._history:
                line += f"User: {turn['question']}\nAI: {turn['answer']}\n"
        line += "Now, continue the conversation:\n"
        line += prompt
        answer = self.ask(line)
        self.__add_turn(prompt, answer)
        return answer
    
    def _clear_history(self) -> None:
        if self._history:
            self._history.clear()
    
    def __enforce_type(self, value, expected_type, arg_name):
        if not isinstance(value, expected_type):
            raise TypeError(f"Argument '{arg_name}' must be of type {expected_type.__name__}, but received {type(value).__name__}")
    
    def __add_turn(self, question, answer) -> None:
        self.__enforce_type(question, str, "question")
        self.__enforce_type(answer, str, answer)

        while len(self._history) >= self._max_length: #change from if to while because the memories could now be modified
            self._history.pop(0)  # Remove the oldest turn
        self._history.append({"question": question, "answer": answer})


from googlesearch import search #need pip install googlesearch-python
import requests #need pip install requests
from bs4 import BeautifulSoup # need pip install beautifulsoup4
from readability import Document # need pip install readability-lxml
import logging

class GoogleSearcher:
    def __init__(self, user_agent: str = "Mozilla/5.0", num_results: int = 5, limit_word_per_url : int = 500):
        self.__enforce_type(user_agent, str, "user_agent")
        self.__enforce_type(num_results, int, "num_results")
        self.__enforce_type(limit_word_per_url, int, "limit_word_per_url")
        self.user_agent = user_agent
        self.num_results = num_results
        self.limit_word = limit_word_per_url
        #for logging only this class rather
        self.logger = logging.getLogger(self.__class__.__name__)
        handler = logging.StreamHandler()
        handler.setFormatter(logging.Formatter(
            '[%(asctime)s] - [%(name)s] - [%(levelname)s] - %(message)s',
            datefmt='%Y-%m-%d %H:%M:%S'
        ))
        self.logger.addHandler(handler)
        self.logger.setLevel(logging.INFO)
        
    def _url_search(self, query) -> list:
        #return the first num_results url
        self.logger.debug("Trying to return the list of urls")
        return list(search(query, num_results= self.num_results))

    def _fetch_html(self, url : str) -> str:
        headers = {"User-Agent": self.user_agent}  #mimic a real browser
        resp = requests.get(url, headers=headers)  # GET the page
        resp.raise_for_status()
        self.logger.debug("Fetching html completed")
        return resp.text

    def _extract_with_readability(self, html : str) -> str:
        doc = Document(html)          # build readability DOM
        summary_html = doc.summary()  # HTML content
        # BeautifulSoup to get plain text
        self.logger.debug("Trying to get the text from html")
        return BeautifulSoup(summary_html, "html.parser").get_text()

    def search(self, prompt : str = None) -> str:
        self.__enforce_type(prompt, str, "prompt")
        
        answer = ""
        urls = self._url_search(prompt)
        for url in urls:
            try:
                result = self._fetch_html(url)
                text = self._extract_with_readability(result)
                answer += text[:self.limit_word]
            except Exception:
                pass #ignore
        
        return answer.replace("\n", " ") #this to reduce the token use if calls API
    
    def __enforce_type(self, value, expected_type, arg_name):
        if not isinstance(value, expected_type):
            raise TypeError(f"Argument '{arg_name}' must be of type {expected_type.__name__}, but received {type(value).__name__}")


from transformers import AutoProcessor, AutoModelForVision2Seq # need pip install transformers 
from PIL import Image # need pip install torch
import torch # need pip install torch
import logging
from typing import Optional
logging.basicConfig(level=logging.INFO, format='[%(asctime)s] - [%(name)s] - [%(levelname)s] - %(message)s', datefmt='%Y-%m-%d %H:%M:%S')

#Another model_name : "Salesforce/blip-image-captioning-base"

class ImageCaptioner:
    def __init__(self, model_name: str = "Salesforce/blip-image-captioning-large", device: Optional[str] = None):
        #check input type
        self.__enforce_type(model_name, str, "model_name")
        
        #logger 
        self.logger = logging.getLogger(self.__class__.__name__)
        self.logger.info("Note: This class takes an image path as input and generates a caption. It is not designed for question answering.")
        
        #init the var to hold device available
        preferred_devices = []
        
        #try input first
        if device is not None:
            self.__enforce_type(device, str, "device")
            preferred_devices.append(device)
        
        # try cuda second 
        if torch.cuda.is_available() and "cuda" not in preferred_devices:
            preferred_devices.append("cuda")

        # fall back to CPU if not already there
        if "cpu" not in preferred_devices:
            preferred_devices.append("cpu")

        last_err = None
        
        for dev in preferred_devices:
            try:
                self.logger.info(f"Loading '{model_name}' model on {dev}.")
                self._device = dev
                self._processor = AutoProcessor.from_pretrained(model_name, use_fast=True)
                self._model = AutoModelForVision2Seq.from_pretrained(model_name).to(self._device)
                self._model.eval()
                self.logger.info(f"Successfully loaded on {dev}.")
                break
            except RuntimeError as e:
                last_err = e
                self.logger.error(f"Failed to load on {dev}: {e}")
            except Exception as e: 
                last_err = e
                self.logger.error(f"An unexpected error occurred while loading model on {dev}: {e}")
        
        if self._model is None: #check if nothing works, then raise error
            raise RuntimeError(f"Could not load model on any device. Last error:\n{last_err}")
            
    @property
    def device(self):
        return self._device

    @property
    def processor(self):
        return self._processor

    @property
    def model(self):
        return self._model

    def write_caption(self, image_path: str, max_length: int = 100, num_beams: int = 3, early_stopping: bool = True) -> str:
        #check type
        self.__enforce_type(image_path, str, "image_path")
        self.__enforce_type(max_length, int, "max_length")
        self.__enforce_type(num_beams, int, "num_beams")
        self.__enforce_type(early_stopping, bool, "early_stopping")
        
        #check num_beam >= 2
        if num_beams < 2:
            raise ValueError(f"num_beams must be >= 2, but got {num_beams}") #Added the value to the error
        
        # load image
        try:
            image = Image.open(image_path).convert("RGB")
        except FileNotFoundError:
            raise FileNotFoundError(f"Image not found: {image_path}")

        inputs = self._processor(images=image, return_tensors="pt").to(self._device)
        with torch.no_grad():
            outputs = self._model.generate(**inputs, max_length=max_length, num_beams=num_beams, early_stopping=early_stopping)

        # text return
        return self._processor.decode(outputs[0], skip_special_tokens=True)

    def __enforce_type(self, value, expected_type, arg_name):
        if not isinstance(value, expected_type):
            raise TypeError(f"Argument '{arg_name}' must be of type {expected_type.__name__}, but received {type(value).__name__}")

#Example
if __name__ == "__main__":
    _cap = ImageCaptioner(device="cuda")
    _text = _cap.write_caption("img.jpg")
    print("Caption:", _text)




import whisper #need pip install openai-whisper
import soundfile as sf #need pip install souldfile
import numpy as np # need pip install numpy
import librosa #need pip install librosa 
import torch # need pip install torch
from typing import Dict, Any
import logging
from typing import Optional

# 3 function use: transcribe -> return Dict (return everything and you choose which to get)
#                 get_lang_detect -> return str (return the language)
#                 get_translation -> return str (return the translation only, for people who doesnt care what language or anything else)

logging.basicConfig(level=logging.INFO, format='[%(asctime)s] - [%(name)s] - [%(levelname)s] - %(message)s', datefmt='%Y-%m-%d %H:%M:%S')

class OpenAIWhisper:
    def __init__(self, model: str = "medium", sample_rate: int = 16000, device : Optional[str]  = None) -> None:
        #check type
        self.__enforce_type(sample_rate, int, sample_rate)
        self.logger = logging.getLogger(self.__class__.__name__)
        
        # model: Whisper model size ("tiny", "base", "small", "medium", "large")
        self._sample_rate = sample_rate
        self._model = None 
        self._device = None 
        
        #init the var to hold device available
        preferred_devices = []
        
        #try input first
        if device is not None:
            self.__enforce_type(device, str, "device")
            preferred_devices.append(device)
        
        # try cuda second 
        if torch.cuda.is_available() and "cuda" not in preferred_devices:
            preferred_devices.append("cuda")

        # fall back to CPU if not already there
        if "cpu" not in preferred_devices:
            preferred_devices.append("cpu")

        last_err = None
        for dev in preferred_devices:
            try:
                self.logger.info(f"Loading '{model}' model on {dev}...")
                self._model = whisper.load_model(model, device=dev)
                self._device = dev
                self.logger.info(f"Model successfully loaded on {dev}.")
                break
            except RuntimeError as e:
                last_err = e
                self.logger.error(f"Failed to load on {dev}: {e}")
            except Exception as e: 
                last_err = e
                self.logger.error(f"An unexpected error occurred while loading model on {dev}: {e}")
        
        if self._model is None: #check if nothing works, then raise error
            raise RuntimeError(f"Could not load model on any device. Last error:\n{last_err}")

    @property
    def sample_rate(self):
        # Returns the sample rate used by the model.
        return self._sample_rate

    @property
    def model(self):
        # Returns the loaded Whisper model.
        return self._model

    @property
    def device(self):
        # Returns the device the model is loaded on.
        return self._device
    
    def _load_wav(self, path: str) -> np.ndarray:
        try:
            audio, orig_sr = sf.read(path)
        except FileNotFoundError:
            self.logger.error(f"Error: Audio file not found at {path}")
            raise
        except Exception as e:
            self.logger.error(f"Error reading audio file {path}: {e}")
            raise

        if audio.ndim > 1:
            audio = np.mean(audio, axis=1) # Change to mono
        
        if orig_sr != self._sample_rate:
            audio = librosa.resample(audio, orig_sr=orig_sr, target_sr=self._sample_rate)
        
        return audio.astype(np.float32)

    def transcribe(self, audio_path: str, fp16: bool = False, **transcribe_kwargs: Any) -> Dict[str, Any]:
        # Transcribes audio from a file path.
        # param audio_path: Path to the audio file.
        # param fp16: Whether to use fp16 precision (GPU only).
        # param transcribe_kwargs: Additional arguments dokter for whisper.model.transcribe().
        # return: A dictionary containing the transcription results.
        
        self.__enforce_type(audio_path, str, "audio_path")
        self.__enforce_type(fp16, bool, "fp16")
        
        try:
            audio_data = self._load_wav(audio_path)
                
            # get the np.ndarray to the transcribe
            self.logger.info(f"Transcribing audio from {audio_path} on {self._device}...")
            result = self._model.transcribe(audio_data, fp16=fp16 if self._device == "cuda" else False, **transcribe_kwargs)
            self.logger.info(f"Transcription successful.")
            return result #return result
        except Exception as e:
            self.logger.error(f"Transcription failed for {audio_path}: {e}")
            raise
    
    def get_translation(self, audio_path: str, fp16: bool = False, **transcribe_kwargs: Any) -> str:
        text = self.transcribe(audio_path=audio_path, fp16=fp16, **transcribe_kwargs)["text"]
        return text

    def get_lang_detect(self, audio_path: str, fp16: bool = False, **transcribe_kwargs: Any) -> str:
        language = self.transcribe(audio_path=audio_path, fp16=fp16, **transcribe_kwargs)["language"]
        return language
    
    def __enforce_type(self, value, expected_type, arg_name):
        if not isinstance(value, expected_type):
            raise TypeError(f"Argument '{arg_name}' must be of type {expected_type.__name__}, but received {type(value).__name__}")
        
# Example usage      
if __name__ == "__main__":
    
    FILENAME = "output_pyaudio.wav"

    _transcriber = OpenAIWhisper()
    _output = _transcriber.transcribe(FILENAME)
    
    #usually, we only care about 2 thing, the transcribe and the language detect
    print("Detected language:", _output["language"])
    print("Transcript:", _output["text"])
    
    #everything in the output for anyone cares
    print(_output)
    
    _language_dect =_transcriber.get_lang_detect(FILENAME)
    print("Language used: ")
    print(_language_dect)
    
    _text_translated = _transcriber.get_translation(FILENAME)
    print("Text transcription only: ")
    print(_text_translated)



import pyttsx3 #need pip install pyttsx3

class Text_To_Speech_Pyttsx3:
    def __init__(self, rate=170, volume=1.0, voice_index=0) -> None:
        self._engine = pyttsx3.init()
        self.config_voice(rate, volume, voice_index)

    def config_voice(self, rate: float = 170, volume: float = 1.0, voice_index: int = 0) -> None:
        if not isinstance(rate, (int, float)):
            raise TypeError("Rate must be a number.")
        if not (0.0 <= volume <= 1.0):
            raise ValueError("Volume must be between 0.0 and 1.0.")
        
        voices = self._engine.getProperty('voices')
        if not isinstance(voice_index, int) or not (0 <= voice_index < len(voices)):
            raise ValueError(f"Voice index must be between 0 and {len(voices)-1}")
        
        self._engine.setProperty('rate', rate)
        self._engine.setProperty('volume', volume)
        self._engine.setProperty('voice', voices[voice_index].id)

    def speak(self, text: str) -> None:
        if not isinstance(text, str):
            raise ValueError(f"The provided text is not string!")
        self._engine.say((text))
        self._engine.runAndWait()

    def stop(self) -> None:
        self._engine.stop()


import os
import fitz # pip install pymupdf
from pypdf import PdfReader # need pip install pypdf
import pdfplumber # need pip install pdfplumber
from typing import Optional
import logging
try:
    from docx import Document # need pip install python-docx
except ImportError:
    Document = None

root_logger = logging.getLogger()
root_logger.setLevel(logging.CRITICAL)

class PDF_DOCX_Reader:
    def __init__(self, start_page: int = 0, last_page: Optional[int] = None) -> None:
        # Initialize with optional page range.
        # Supports PDF and DOCX formats.
        # param first_page: zero-based index of first page to process
        # param last_page: zero-based index of last page, None means all pages
        self.__enforce_type(start_page, int, "first_page")
        self.__enforce_type(last_page, (int, type(None)), "last_page")
        
        self.first_page = start_page
        self.last_page = last_page
        #for logging only this class rather
        self.logger = logging.getLogger(self.__class__.__name__)
        self.logger.propagate = False  # Prevent propagation to the root logger
        handler = logging.StreamHandler()
        handler.setFormatter(logging.Formatter(
            '[%(asctime)s] - [%(name)s] - [%(levelname)s] - %(message)s',
            datefmt='%Y-%m-%d %H:%M:%S'
        ))
        self.logger.addHandler(handler)
        self.logger.setLevel(logging.INFO)
        self.logger.info(f"Initialize successfully")

    def extract_all_text(self, file_path: str = None, first_page: Optional[int] = None, last_page: Optional[int] = None) -> str:
        # Extract all text into a single string.
        # Tries pypdf for PDF, Document for DOCX, and falls back to fitz for PDFs.
        
        #check type
        self.__enforce_type(first_page, (int, type(None)), "first_page")
        self.__enforce_type(last_page, (int, type(None)), "last_page")
        
        self.__enforce_type(file_path, str, file_path)
        ext = self.__isSupported(file_path) #extract the end (file type)
        
        #after config, you can still choose the start and end page for specific one
        fp = first_page if first_page is not None else self.first_page
        lp = last_page if last_page is not None else self.last_page

        self.logger.info(f"Detect file type: {ext}")
        
        # DOCX
        if ext == '.docx':
            if Document is None:
                raise ImportError("python-docx is required to read DOCX files.")
            doc = Document(file_path)
            text = []
            try:
                for para in doc.paragraphs:
                    if para.text:
                        text.append(para.text)

                for table in doc.tables:
                    for row in table.rows:
                        for cell in row.cells:
                            if cell.text:
                                text.append(cell.text)
                return '\n'.join(text)
            except Exception as e:
                self.logger.error(f"Fail to get text from file {file_path} with error {e}")
                return ""

        # PDF
        text = []
        try:
            reader = PdfReader(file_path)
            pages = reader.pages
            total = len(pages)
            end = lp + 1 if lp is not None else total
            for i in range(fp, min(end, total)):
                page = pages[i]
                txt = page.extract_text() or ''
                text.append(txt)
            return '\n'.join(text)
           
        except Exception:
            # fallback to fitz if can't use pypdf
            self.logger.error(f"Fail to use pypdf, fallback to fitz to extract file")
            doc = fitz.open(file_path)
            total = doc.page_count
            end = lp + 1 if lp is not None else total
            try:
                for i in range(fp, min(end, total)):
                    page = doc.load_page(i)
                    text.append(page.get_text())
                return '\n'.join(text)
            except Exception as e:
                self.logger.error(f"Fail to get text from file {file_path} with error {e}")
                return ""

    def extract_ordered_text( self, file_path: str = None, first_page: Optional[int] = None, last_page: Optional[int] = None) -> str:
        # Extract text maintaining layout using pdfplumber for PDFs,
        # or fallback to fitz. DOCX behaves same as extract_all_text.
        
        #check type
        self.__enforce_type(first_page, (int, type(None)), "first_page")
        self.__enforce_type(last_page, (int, type(None)), "last_page")

        #path, file checking
        self.__enforce_type(file_path, str, "file_path")
        ext = self.__isSupported(file_path)
        
        fp = first_page if first_page is not None else self.first_page
        lp = last_page if last_page is not None else self.last_page
        
        # DOCX same as extract_all_text
        if ext == '.docx':
            #adjust here the code please
            try:
                from docx.oxml.text.paragraph import CT_P
                from docx.oxml.table import CT_Tbl
                from docx.text.paragraph import Paragraph
                from docx.table import Table
                doc = Document(file_path)
                parts = []
                for child in doc.element.body.iterchildren():
                    if isinstance(child, CT_P):
                        para = Paragraph(child, doc)
                        parts.append(para.text)
                    elif isinstance(child, CT_Tbl):
                        tbl = Table(child, doc)
                        # turn each row into a tab‑joined line
                        for row in tbl.rows:
                            parts.append("\t".join(cell.text for cell in row.cells))
                return "\n".join(parts)
            except Exception:
                self.logger.error("Fail to do extract_ordered_text for docx file, trying extract_all_text instead")
                return self.extract_all_text(file_path, fp, lp)
        
        # PDF using pdfplumber
        text = []
        try:
            with pdfplumber.open(file_path) as pdf:
                total = len(pdf.pages)
                end = lp + 1 if lp is not None else total
                for i in range(fp, min(end, total)):
                    page = pdf.pages[i]
                    # preserve layout spacing
                    txt = page.extract_text(x_tolerance=2, y_tolerance=2) or ''
                    text.append(txt)
            return '\n'.join(text)
        except Exception:
            # fallback to fitz
            self.logger.info(f"Fail to use pdflumber, fallback to fitz to extract file")
            return self.extract_all_text(file_path, fp, lp)

    def extract_images(self, file_path: str = None, folder_extract: str = "extracted_images", first_page: Optional[int] = None, last_page: Optional[int] = None) -> int:
        # Extract images from PDF using fitz (PyMuPDF).
        # Saves images to folder_extract, and returns count.
   
        #check type
        self.__enforce_type(first_page, (int, type(None)), "first_page")
        self.__enforce_type(last_page, (int, type(None)), "last_page")

        #path, file checking
        self.__enforce_type(file_path, str, "file_path")
        ext = self.__isSupported(file_path)
        
        fp = first_page if first_page is not None else self.first_page
        lp = last_page if last_page is not None else self.last_page

        os.makedirs(folder_extract, exist_ok=True) #folder create or use existence one
        count = 0

        self.logger.info(f"Detect file type: {ext}")
        
        if ext == '.pdf':
            doc = fitz.open(file_path)
            total = doc.page_count
            end = lp + 1 if lp is not None else total
            for i in range(fp, min(end, total)):
                page = doc.load_page(i)
                images = page.get_images(full=True)
                for img_index, img in enumerate(images):
                    xref = img[0]
                    base_image = doc.extract_image(xref)
                    img_bytes = base_image["image"]
                    extn = base_image.get("ext", "png")
                    img_name = f"page{i+1}_img{img_index+1}.{extn}"
                    out_path = os.path.join(folder_extract, img_name)
                    with open(out_path, 'wb') as f:
                        f.write(img_bytes)
                    count += 1
        else:
            # DOCX image extraction using related_parts
            if Document is None:
                raise ImportError("python-docx is required to extract images from DOCX files.")
            doc = Document(file_path)
            # collect all image parts
            image_parts = [part for part in doc.part.related_parts.values()
                           if hasattr(part, 'content_type') and part.content_type.startswith('image/')]
            for part in image_parts:
                img_bytes = part.blob
                img_name = os.path.basename(part.partname)
                out_path = os.path.join(folder_extract, img_name)
                with open(out_path, 'wb') as f:
                    f.write(img_bytes)
                count += 1

        self.logger.info(f"Complete extracted images in file {file_path} to folder {folder_extract}")
        return count #return number of images found in the file

    def __isSupported(self, file_path : str) -> str:
        if file_path:
            ext = os.path.splitext(file_path)[1].lower()
            if ext not in ['.pdf', '.docx']:
                raise ValueError(f"Unsupported file type: {ext}")
            return ext
    
    def __enforce_type(self, value, expected_types, arg_name) -> None:
        if not isinstance(value, expected_types):
            expected_names = [t.__name__ for t in expected_types] if isinstance(expected_types, tuple) else [expected_types.__name__]
            expected_str = ", ".join(expected_names)
            raise TypeError(f"Argument '{arg_name}' must be of type {expected_str}, but received {type(value).__name__}")


import pyaudio # need pip install pyaudio
import wave
import time
import numpy as np # need pip install numpy
import keyboard #need pip install keyboard
from lameenc import Encoder #need pip install lameenc (this is use for MP3Recorder)
import logging
import os

#Record with pyaudio
#3 function: fixed-duration recording, toggle-controlled recording, and silence-based recording.

logging.basicConfig(level=logging.INFO, format='[%(asctime)s] - [%(name)s] - [%(levelname)s] - %(message)s', datefmt='%Y-%m-%d %H:%M:%S')

class WavRecorder:
    def __init__(self, channels=1, rate=44100, chunk=1024, fmt=pyaudio.paInt16):
        self._channels = channels
        self._rate = rate
        self._chunk = chunk
        self._format = fmt
        self._p = pyaudio.PyAudio()
        self._frames = []
        self._stream = None
        self._recording = False
        self.logger = logging.getLogger(self.__class__.__name__)
        self.logger.info("NOTE: WavRecorder supports only .wav output files.")
    
    @property
    def channels(self):
        # Returns the number of audio channels.
        return self._channels

    @property
    def rate(self):
        # Returns the audio sample rate.
        return self._rate

    @property
    def chunk(self):
        # Returns the audio chunk size.
        return self._chunk

    @property
    def format(self):
        # Returns the audio format.
        return self._format
       
    def _open_stream(self):
        self._stream = self._p.open(format=self._format, channels=self._channels, rate=self._rate, input=True, frames_per_buffer=self._chunk)
        self._frames = []

    def _close_stream(self):
        if self._stream is not None:
            self._stream.stop_stream()
            self._stream.close()
            self._stream = None

    def _save_wave(self, filename):
        self.logger.debug(f"Tryint to save audio file as {filename}")
        wf = wave.open(filename, 'wb')
        wf.setnchannels(self._channels)
        wf.setsampwidth(self._p.get_sample_size(self._format))
        wf.setframerate(self._rate)
        wf.writeframes(b''.join(self._frames))
        wf.close()
        self.logger.info(f"Audio saved as {filename}")

    def __rms(self, data):
        #switch from audioop to this because python 3.13 wont support it anymore
        samples = np.frombuffer(data, dtype=np.int16).astype(np.float64)
        return np.sqrt(np.mean(samples**2))

    def record_fixed(self, duration : int, output_filename : str ="fixed_record.wav") -> None:  # duration in seconds
        self.__enforce_type(duration, int, "duration")
        self.__enforce_type(output_filename, str, "output_filename")
        
        # record audio for a fixed duration
        self._open_stream()
        self.logger.info(f"Recording for {duration} seconds...")
        
        try:
            for _ in range(0, int(self._rate / self._chunk * duration)):
                data = self._stream.read(self._chunk)
                self._frames.append(data)
        except KeyboardInterrupt:
            self.logger.critical("Recording stopped manually using CTRL + C .")
            raise
        except Exception as e:
            self.logger.error(f"An unexpected error occurred: {e}")
            raise
        
        finally:
            self.logger.info("Fixed-time recording finished.")
            self._close_stream()
            self._save_wave(output_filename)

    def record_toggle(self, toggle_key : str ='`', output_filename : str ="toggle_record.wav") -> None:  # default toggle key: backtick
        #check type
        self.__enforce_type(toggle_key, str, "toggle_key")
        self.__enforce_type(output_filename, str, "output_filename")
        
        #Start recording until toggle_key is pressed again.
        self._open_stream()
        self._recording = True
        self.logger.info(f"Recording... press '{toggle_key}' to stop.")

        # Hotkey to toggle recording state
        keyboard.add_hotkey(toggle_key, lambda: setattr(self, '_recording', False))
        try:
            while self._recording:
                data = self._stream.read(self._chunk)
                self._frames.append(data)
                # slight sleep to allow key event processing
                time.sleep(0.01)
        except KeyboardInterrupt:
            self.logger.critical("Recording stopped manually using CTRL + C.")
            raise
        except Exception as e:
            self.logger.error(f"An unexpected error occurred: {e}")
            raise
        
        finally:    
            self.logger.info("Toggle recording stopped.")
            keyboard.remove_hotkey(toggle_key)
            self._close_stream()
            self._save_wave(output_filename)

    def record_silence(self, silence_threshold : int = 800, max_silence_seconds : int = 3, output_filename : str ="silence_record.wav") -> None: 
        #check type 
        self.__enforce_type(silence_threshold, int, "silence_threshold")
        self.__enforce_type(max_silence_seconds, int, "max_silence_seconds")
        self.__enforce_type(output_filename, str, "output_filename")
        
        #Record until a period of silence longer than max_silence_seconds is detected.
        self._open_stream()
        self.logger.info(f"Recording... will stop after silence of > {max_silence_seconds}s below threshold {silence_threshold}.")
        silence_start = None

        try:
            while True:
                data = self._stream.read(self._chunk)
                self._frames.append(data)

                rms_val = self.__rms(data)
                # print(rms_val) # check value it is hearing so you could adjust base on your need
                if rms_val < silence_threshold:
                    if silence_start is None:
                        silence_start = time.time()
                    elif time.time() - silence_start > max_silence_seconds:
                        self.logger.info("Silence detected. Stopping...")
                        break
                else:
                    silence_start = None
        except KeyboardInterrupt:
            self.logger.critical("Recording stopped manually using CTRL + C.")
            raise
        except Exception as e:
            self.logger.error(f"An unexpected error occurred: {e}")
            raise
        
        finally:
            self._close_stream()
            self._save_wave(output_filename)

    def terminate(self):
        self._p.terminate()

    def __enforce_type(self, value, expected_type, arg_name):
        if not isinstance(value, expected_type):
            raise TypeError(f"Argument '{arg_name}' must be of type {expected_type.__name__}, but received {type(value).__name__}")

class MP3Recorder:
    def __init__(self, channels=1, rate=44100, chunk=1024, bitrate=192):
        self._channels = channels
        self._rate = rate
        self._chunk = chunk
        self._bitrate = bitrate

        # Initialize lame encoder
        self.encoder = Encoder()
        self.encoder.set_bit_rate(self._bitrate)
        self.encoder.set_in_sample_rate(self._rate)
        self.encoder.set_channels(self._channels)
        self.encoder.set_quality(2)  # 2 = high quality

        self._p = pyaudio.PyAudio()
        self._frames = []  # raw PCM frames
        self._stream = None
        self._recording = False
        self.logger = logging.getLogger(self.__class__.__name__)
        self.logger.info("NOTE: MP3Recorder supports only .mp3 output files.")

    @property
    def channels(self):
        # Returns the number of audio channels.
        return self._channels

    @property
    def rate(self):
        # Returns the audio sample rate.
        return self._rate

    @property
    def chunk(self):
        # Returns the audio chunk size.
        return self._chunk

    @property
    def bitrate(self):
        # Returns the audio bitrate.
        return self._bitrate

    def _open_stream(self):
        self._stream = self._p.open(
            format=pyaudio.paInt16,
            channels=self._channels,
            rate=self._rate,
            input=True,
            frames_per_buffer=self._chunk
        )
        self._frames = []

    def _close_stream(self):
        if self._stream:
            self._stream.stop_stream()
            self._stream.close()
            self._stream = None

    def _save_mp3(self, filename):
        # Encode raw PCM to MP3 and write
        self.logger.debug(f"Tryint to save audio file as {filename}")
        with open(filename, 'wb') as f:
            for chunk in self._frames:
                mp3_data = self.encoder.encode(chunk)
                if mp3_data:
                    f.write(mp3_data)
            # flush encoder
            f.write(self.encoder.flush())
        self.logger.info(f"Audio saved as {filename}")

    def __rms(self, data):
        samples = np.frombuffer(data, dtype=np.int16).astype(np.float64)
        return np.sqrt(np.mean(samples**2))

    def __enforce_type(self, value, expected_type, arg_name):
        if not isinstance(value, expected_type):
            raise TypeError(
                f"Argument '{arg_name}' must be of type {expected_type.__name__}, but received {type(value).__name__}")

    def record_fixed(self, duration: int, output_filename: str ="fixed_record.mp3") -> None:
        self.__enforce_type(duration, int, "duration")
        self.__enforce_type(output_filename, str, "output_filename")

        self._open_stream()
        self.logger.info(f"Recording for {duration} seconds...")
        
        try:
            for _ in range(int(self._rate / self._chunk * duration)):
                data = self._stream.read(self._chunk)
                self._frames.append(data)
        except KeyboardInterrupt:
            self.logger.critical("Recording stopped manually using CTRL + C.")
            raise
        except Exception as e:
            self.logger.error(f"An unexpected error occurred: {e}")
            raise
                    
        finally:        
            self.logger.info("Fixed-time recording finished.")
            self._close_stream()
            self._save_mp3(output_filename)

    def record_toggle(self, toggle_key: str ='`', output_filename: str ="toggle_record.mp3") -> None:
        self.__enforce_type(toggle_key, str, "toggle_key")
        self.__enforce_type(output_filename, str, "output_filename")

        self._open_stream()
        self._recording = True
        self.logger.info(f"Recording... press '{toggle_key}' to stop.")
        keyboard.add_hotkey(toggle_key, lambda: setattr(self, '_recording', False))
        
        try: 
            while self._recording:
                data = self._stream.read(self._chunk)
                self._frames.append(data)
                time.sleep(0.01)
        except KeyboardInterrupt:
            self.logger.critical("Recording stopped manually using CTRL + C.")
            raise
        except Exception as e:
            self.logger.error(f"An unexpected error occurred: {e}")
            raise
            
        finally:
            self.logger.info("Toggle recording stopped.")
            keyboard.remove_hotkey(toggle_key)
            self._close_stream()
            self._save_mp3(output_filename)

    def record_silence(self, silence_threshold: int = 800, max_silence_seconds: int = 3, output_filename: str ="silence_record.mp3") -> None:
        self.__enforce_type(silence_threshold, int, "silence_threshold")
        self.__enforce_type(max_silence_seconds, int, "max_silence_seconds")
        self.__enforce_type(output_filename, str, "output_filename")

        self._open_stream()
        self.logger.info(f"Recording... will stop after >{max_silence_seconds}s below threshold {silence_threshold}.")
        silence_start = None
        try:
            while True:
                data = self._stream.read(self._chunk)
                self._frames.append(data)
                rms_val = self.__rms(data)
                if rms_val < silence_threshold:
                    if silence_start is None:
                        silence_start = time.time()
                    elif time.time() - silence_start > max_silence_seconds:
                        self.logger.info("Silence detected. Stopping...")
                        break
                else:
                    silence_start = None
        except KeyboardInterrupt:
            self.logger.critical("Recording stopped manually using CTRL + C.")
            raise
        except Exception as e:
            self.logger.error(f"An unexpected error occurred: {e}")
            raise
            
        finally:
            self._close_stream()
            self._save_mp3(output_filename)

    def terminate(self):
        self._p.terminate()

from mutagen.mp3 import MP3, MutagenError

def check_wav_length_and_size(file_path: str, period: float = 5.0, allowed_logger : bool = True) -> bool:
    logger = logging.getLogger("WAV Checker")
    if allowed_logger:
        logger.setLevel(logging.INFO)
    else: logger.setLevel(logging.ERROR)
    try:
        with wave.open(file_path, 'rb') as wf:
            frame_rate = wf.getframerate()
            num_frames = wf.getnframes()
            duration = num_frames / float(frame_rate)

        file_size_bytes = os.path.getsize(file_path)

        if duration <= period:
            logger.info("'%s' is less than or equal to %.2f seconds long (duration: %.2f seconds).", 
                        file_path, period, duration)
            logger.info("File size: %d bytes.", file_size_bytes)
            return True
        else:
            logger.info("'%s' is %.2f seconds long, which exceeds %.2f seconds.", 
                           file_path, duration, period)
            return False
    except wave.Error as e:
        logger.error(f"Error opening or reading WAV file '{file_path}': {e}")
        raise
    except FileNotFoundError:
        logger.error(f"Error: File not found at '{file_path}'")
        raise
    except Exception as e:
        logger.error(f"An unexpected error occurred: {e}")
        raise

def check_mp3_length_and_size(file_path: str, period: float = 5.0, allowed_logger : bool = True) -> bool:
    logger = logging.getLogger("MP3 Checker")
    if allowed_logger:
        logger.setLevel(logging.INFO)
    else: logger.setLevel(logging.ERROR)
    try:
        audio = MP3(file_path)
        duration = audio.info.length  # length in seconds
        file_size_bytes = os.path.getsize(file_path)

        if duration <= period:
            logger.info("'%s' is less than or equal to %.2f seconds long (duration: %.2f seconds).", 
                        file_path, period, duration)
            logger.info("File size: %d bytes.", file_size_bytes)
            return True
        else:
            logger.info("'%s' is %.2f seconds long, which exceeds %.2f seconds.", 
                           file_path, duration, period)
            return False
    except MutagenError as e:
        logger.error("Mutagen error reading MP3 file '%s': %s", file_path, e)
        raise
    except FileNotFoundError:
        logger.error(f"Error: File not found at '{file_path}'")
        raise
    except Exception as e:
        logger.error("Error processing MP3 file '%s': %s", file_path, e)
        raise

#EXAMPLE
def _test_function():
    _rec = WavRecorder()
    _rec.record_toggle()
        
if __name__ == "__main__":
    import multiprocessing
    _p = multiprocessing.Process(target=_test_function)
    _p.start()
    # Simulate later termination, this would not provide you any audio output
    time.sleep(5)
    _p.terminate()  
    _p.join()
    
    _d = multiprocessing.Process(target=function) #this would pause until you actually press ` to stop recording
    _d.start()
    _d.join()
    
    _rec = WavRecorder() # you could just use as normal if you dont need to interrupt midway 
    _rec.record_fixed(3) 
    
    _rec = MP3Recorder()
    _rec.record_fixed(4)
    


import time
from functools import wraps

#  A decorator that measures and prints the execution time of a function.
def time_it(func):
    @wraps(func) 
    def wrapper(*args, **kwargs):
        start_time = time.perf_counter()
        result = func(*args, **kwargs)
        end_time = time.perf_counter()
        execution_time = end_time - start_time
        print(f"Function '{func.__name__}' executed in {execution_time:.4f} seconds.")
        return result
    return wrapper

#how to use
if __name__== "__main__":
    @time_it
    def say_hello():
        print("Hello world!")  
    say_hello()